#include "anircd.h"

Chan *clist = NULL;
User *ulist = NULL;

int g_numchans = 0, g_numusers = 0;

Chan *chan_new(char *name)
{
	Chan	*c;

	c = (Chan *) malloc(sizeof(Chan));
	if (!c)
	{
		log_error("Failed to allocate Chan struct in %s!", __func__);
		return(NULL);
	}

	c->name = strdup(name);
	c->nusers = 0;
	c->users = NULL;

	c->next = clist;
	c->prev = NULL;

	if (clist)
		clist->prev = c;

	clist = c;

	++g_numchans;
	return(c);
}

int chan_del(Chan *c)
{
	if (!c)
	{
		log_error("%s called with null user struct!", __func__);
		return(0);
	}

	if (c->nusers)
	{
		log_error("%s called with users still in-channel! Not deleting channel; check callers.", __func__);
		return(0);
	}

	free(c->name);
	/* XXX: Member structs already freed -- we assume we're being called after all users have left. Be sure of caller! */

	if (c->prev)
		c->prev->next = c->next;
	else
		clist = c->next;
	if (c->next)
		c->next->prev = c->prev;

	free(c);
	--g_numchans;
	return(1);
}

Chan *chan_find(char *name)
{
	Chan	*c;

	if (!name)
	{
		log_error("%s called with NULL name parameter", __func__);
		return(NULL);
	}

	for (c = clist; c; c = c->next)
		if (!strcasecmp(c->name, name))
			return(c);

	return(NULL);
}

int chan_finduidx(User *u, Chan *c)
{
	int	i;

	for (i = 0; i < c->nusers; ++i)
		if (c->users[i] == u)
			return(i);

	return(-1);
}

int chan_adduent(User *u, Chan *c)
{
	if (!u || !c)
	{
		log_error("NULL u/c parameters passed to %s!", __func__);
		return(0);
	}

	if (chan_finduidx(u, c) >= 0)
	{
		log_error("Duplicate call to %s - user already in channel", __func__);
		return(0);
	}

	if (user_findcidx(u, c) >= 0)
	{
		log_error("Desynchronisation between u/c structs in %s: user not on c->users, but u->chans?!", __func__);
		return(0);
	}

	c->users = (User **) realloc(c->users, ++c->nusers * sizeof(User *));
	c->users[c->nusers - 1] = u;

	u->chans = (Chan **) realloc(u->chans, ++u->nchans * sizeof(Chan *));
	u->chans[u->nchans - 1] = c;

	return(1);
}

int chan_deluent(User *u, Chan *c)
{
	int	idx;

	idx = chan_finduidx(u, c);
	if (idx < 0)
		return(0);

	if (c->nusers == 1)
	{
		free(c->users);
		c->nusers = 0;
		c->users = NULL;
		chan_del(c);
	}
	else
	{
		if (idx != c->nusers - 1)
			memmove(&c->users[idx], &c->users[idx+1], (c->nusers - (idx + 1)) * sizeof(User *));
		c->users = (User **) realloc(c->users, sizeof(User *) * --c->nusers);
	}

	idx = user_findcidx(u, c);
	if (idx < 0)
		return(0);

	if (u->nchans == 1)
	{
		free(u->chans);
		u->nchans = 0;
		u->chans = NULL;
	}
	else
	{
		if (idx != u->nchans - 1)
			memmove(&u->chans[idx], &u->chans[idx+1], (u->nchans - (idx + 1)) * sizeof(Chan *));
		u->chans = (Chan **) realloc(u->chans, sizeof(Chan *) * --u->nchans);
	}

	return(1);
}

User *user_new(int s)
{
	User	*u;

	u = (User *) malloc(sizeof(User));
	if (!u)
	{
		log_error("Failed to allocate new User structure in %s: %s", __func__, strerror(errno));
		return(NULL);
	}

	u->next = ulist;
	u->prev = NULL;

	u->s = s;
	u->ssl = NULL;

	u->nick = NULL;
	u->realname = NULL;
	u->host = NULL;
	u->ident = NULL;
	u->signon = 0;	

	u->chans = NULL;
	u->nchans = 0;
	u->flags = 0;
	
	u->inbuf = u->outbuf = NULL;
	u->inlen = 0;

	if (ulist)
		ulist->prev = u;

	ulist = u;

	++g_numusers;
	return(u);
}

int user_del(User *u)
{
	Chan	*c;
	int	i;

	if (!u)
	{
		log_error("%s called with null User struct!", __func__);
		return(0);
	}

	if (u->nchans)
		while (u->chans)
			chan_deluent(u, u->chans[0]);

	if (u->nick)
		free(u->nick);
	if (u->realname)
		free(u->realname);
	if (u->host)
		free(u->host);
	if (u->ident)
		free(u->ident);

	if (u->inbuf)
		free(u->inbuf);
	if (u->outbuf)
		free(u->outbuf);

	if (u->prev)
		u->prev->next = u->next;
	else
		ulist = u->next;

	if (u->next)
		u->next->prev = u->prev;
	free(u);

	--g_numusers;
	return(1);
}

User *user_find(char *nick)
{
	User	*u;

	if (!nick)
	{
		log_error("%s called with NULL nick parameter", __func__);
		return(NULL);
	}

	for (u = ulist; u; u = u->next)
		if (u->nick && !strcasecmp(u->nick, nick))
			return(u);

	return(NULL);
}

int user_findcidx(User *u, Chan *c)
{
	int	i;

	for (i = 0; i < u->nchans; ++i)
		if (u->chans[i] == c)
			return(i);

	return(-1);
}

void user_kill(User *u, char *message)
{
	send_cmd_nobuf(u, NULL, "ERROR :Closing Link: %s (%s)", u->nick ? u->nick : "*", message);
	user_kill2(u, message);
}

void user_kill2(User *u, char *message)
{
	send_ucmd_chanbutone(u, "QUIT :%s", message);
	SSL_shutdown(u->ssl);
	SSL_free(u->ssl);

	close(u->s);
	user_del(u);
}
